using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;


namespace Zcu.Gsvd.Extractions
{
    /// <summary>
    /// <b>Centred Lattice Cube</b> is method for iso-surface extraction. Each cell is split into 24 tetraherons
    /// whereas tetrahedrons trench into neighbouring cell. All tetraherons (24) are same shape. In comparsion
    /// with Marching Cubes is CCL: generate smooth surface, more computing time expensive, much more tringles 
    /// in output grid and values in the middle of cells are approximated.
    /// </summary>
    [ModuleInfo("Novak Frantisek", "Centred Lattice Cube iso-surface extractor.", IconName = "CCL.ico", Category = "Gsvd")]
    [PortInfo("RegGridND", "RegGridND")]
    [PortInfo("Treshold", "Treshold")]
    [PortInfo("Out", "TriangleMesh")]
    public class CCL : Zcu.Mve.Core.Module
    {
        #region Variables declaration

        /// <summary>
        /// Treshold value in %
        /// </summary>
        public static double TresholdValue;
        private double tresholdPercent = 50;
        private bool useThresholdAsIsoValue = false;
        private Point3D delta; // distance between values in grid      
        private Point3D halfDelta;
        private Point3D distortion;
        private Octahedron oHedron = null; // 4 tetrahedrons with shared edges

        private int kdCells;
        private int indicesValue;
        private List<int> indices = null;
        private List<Point3D> vertices = null;
        private List<Normal3D> normals = null;
        private KdCell[] kdCell = null; // kd-tree cells
        private KdNode[] kdNode = null; // kd-tree nodes

        private int row;   // X
        private int slice; // X * Y
        private int length; // X * Y * Z
        private double min = double.MaxValue, max = double.MinValue;        
        private int[] size; // x, y, z
        private RegGridND data = null; // input data
        private MveArray vData = null;  // volumetric data
        private Dictionary<DEdge, int> dictionary = null;

        private Cell[] upBuffer = null;     // global buffer
        private Cell[] middleBuffer = null; // global buffer
        private Cell[] downBuffer = null;   // global buffer
        private Edge[] localBuffer = null;  // local buffer

        private bool paramsChanged;
        private string pointAttributesName = "Values";
        private string normalAttributeName = "Normal3D";
        private bool kdTree;
        private bool kdBuild;

        private TriangleMesh triMesh = null;

        #endregion

        /// <summary>
        /// Default constructor.
        /// </summary>
        public CCL()
        {
            AddInPort("RegGridND", typeof(RegGridND));
            AddInPort("Treshold", typeof(Integer), false);
            AddOutPort("Out", typeof(TriangleMesh));

            this.paramsChanged = true;
            this.kdTree = false;
        }

        /// <summary>
        /// Entry point of module. By parameters choose CCL with or without preprocessing.
        /// </summary>
        public override void Execute()
        {
            if (!IsAllDataSame() || this.paramsChanged)
            {
                long dt;
                System.Diagnostics.Debug.WriteLine("Centred Lattice Cube");

                System.GC.Collect();

                this.data = (RegGridND)this.GetInput("RegGridND");
                this.vData = data.GetPointAttrs(this.pointAttributesName);

                if (this.GetInput("Treshold") != null)
                    this.tresholdPercent = (Integer)this.GetInput("Treshold");

                if (!IsDataSame("RegGridND"))
                    this.kdBuild = false;

                if (this.kdTree) //kd-tree preprocessing
                {
                    this.indicesValue = 0;

                    if (useThresholdAsIsoValue)
                        CCL.TresholdValue = this.tresholdPercent;
                    else
                        CCL.TresholdValue = this.min + (this.max - this.min) * this.tresholdPercent * 0.01f;

                    if (!this.kdBuild)
                    {
                        this.kdCells = 0;
                        Initialize();
                        this.kdCell = new KdCell[this.length / 2];
                        this.kdBuild = true;
                        BuildKdTree();
                    }

                    dt = Environment.TickCount;
                    WalkThroughMinLevel(0, this.kdCells);
                }
                else // CCL without preprocessing
                {
                    Initialize();
                    dt = Environment.TickCount;
                    WalkThrough();
                }

                this.vData = null;

                System.Diagnostics.Debug.WriteLine(string.Format(" Treashold: {0}%", this.tresholdPercent));
                System.Diagnostics.Debug.WriteLine(string.Format(" Iso-surface extraction time: {0} ms", Environment.TickCount - dt));
                System.Diagnostics.Debug.WriteLine(string.Format(" Triangles: {0}", this.indices.Count / 3, this.vertices.Count));
                System.Diagnostics.Debug.WriteLine(string.Format(" Points: {0}\n", this.vertices.Count));

                CreateMesh();

                SetOutput("Out", this.triMesh);

                this.paramsChanged = false;
                System.GC.Collect();
            }
            else
                SetOutput("Out", this.triMesh, true);
        }

        /// <summary>
        /// Fill triangleMesh with points, normals and trinagles.
        /// </summary>
        private void CreateMesh()
        {
            this.triMesh = new TriangleMesh();

            triMesh.Points = UniformDataArray.FromList<Point3D>(this.vertices);
            this.vertices.Clear();

            ProgressInfo(0.86f);

            UniformDataArray cells = new UniformDataArray(typeof(Triangle), this.indices.Count / 3);
            int index = 0;
            for (int i = 0; i < this.indices.Count; i += 3)
            {
                cells[index] = new Triangle(indices[i], indices[i + 1], indices[i + 2]);
                index++;
            }

            this.triMesh.AddCells("Triangles", cells);
            this.indices.Clear();

            ProgressInfo(0.93f);

            UniformDataArray normals = new UniformDataArray(typeof(Normal3D), this.normals.Count);
            for (int i = 0; i < this.normals.Count; i++)
                normals[i] = this.normals[i];

            this.triMesh.AddPointAttrs(this.normalAttributeName, UniformDataArray.FromList<Normal3D>(this.normals));

            this.normals.Clear();

            ProgressInfo(1f);
        }

        /// <summary>
        /// Inicialization structures.
        /// </summary>
        private void Initialize()
        {
            this.size = this.vData.Lengths;
            this.length = this.size[0] * this.size[1] * this.size[2];

            this.indices = new List<int>(this.length); // triangle indices
            this.vertices = new List<Point3D>(this.length / 2); // points
            this.normals = new List<Normal3D>(this.length / 2); //normals

            this.slice = this.size[0] * this.size[1];
            this.row = this.size[0];
            this.indicesValue = 0;

            // gloabal buffers
            this.upBuffer = new Cell[this.slice];
            this.middleBuffer = new Cell[this.slice];
            this.downBuffer = new Cell[this.slice];
            this.dictionary = new Dictionary<DEdge, int>(this.slice * 2);

            // loacal buffer, used during coumputing one cell
            this.localBuffer = new Edge[20]; // shared edges
            this.oHedron = new Octahedron();

            this.delta = new Point3D(1, 1, 1);
            this.halfDelta = new Point3D(0.5f, 0.5f, 0.5f);
            this.distortion = new Point3D(1, 1, 1);

            bool allAtributes = true;

            if (data.UserTags.ContainsKey("Min"))
                this.min = (Scalar)data.UserTags["Min"];
            else
            {
                this.min = 0;
                allAtributes = false;
            }

            if (data.UserTags.ContainsKey("Max"))
                this.max = (Scalar)data.UserTags["Max"];
            else
            {
                this.max = 255;
                allAtributes = false;
            }

            if (data.UserTags.ContainsKey("Distortion"))
                this.distortion = (Point3D)data.UserTags["Distortion"];
            else
            {
                distortion = new Point3D(1, 1, 1);
                allAtributes = false;
            }

            this.distortion = new Point3D(1.0 / this.distortion.X, 1.0 / this.distortion.Y, 1.0 / this.distortion.Z);

            if (!allAtributes)
                throw new Exception("Attributes not set. For extraction used default values.");

            if (useThresholdAsIsoValue)
                CCL.TresholdValue = this.tresholdPercent;
            else
                CCL.TresholdValue = this.min + (this.max - this.min) * this.tresholdPercent * 0.01f;
        }

        #region CCL + Kd-tree

        /// <summary>
        /// Method used for recursive building of tree at min level. 
        /// Min level = values are sort by minimal value of cell.
        /// </summary>
        /// <param name="dm">lower limit</param>
        /// <param name="length">number cells</param>
        private void MinLevel(int dm, int length)
        {
            Array.Sort<KdNode>(this.kdNode, dm, length, new SortByMin());
            int root = dm + length / 2;
            int hm = dm + length;

            if (dm < root - 1) MaxLevel(dm, root - dm); // left subtree
            if (hm > root + 1) MaxLevel(root + 1, hm - (root + 1)); // right subtree
        }

        /// <summary>
        /// Method used for recursive building of tree at max level.
        /// Max level = values are sort by maximum value of cell.
        /// </summary>
        /// <param name="dm">upper limit</param>
        /// <param name="length">number cells</param>
        private void MaxLevel(int dm, int length)
        {
            Array.Sort<KdNode>(this.kdNode, dm, length, new SortByMax());
            int root = dm + length / 2;
            int hm = dm + length;

            if (dm < root - 1) MinLevel(dm, root - dm); // left subtree
            if (hm > root + 1) MinLevel(root + 1, hm - (root + 1)); // right subtree
        }

        /// <summary>
        /// Passing through the tree at min level and searching active cells.
        /// </summary>
        /// <param name="dm">lower limit</param>
        /// <param name="length">umber cells</param>
        private void WalkThroughMinLevel(int dm, int length)
        {
            int root = dm + length / 2;
            int hm = dm + length;

            // processing of root
            if (kdNode[root].Min <= CCL.TresholdValue && kdNode[root].Max >= CCL.TresholdValue)
                Tetrahedronize(kdCell[kdNode[root].pKdCell]);

            if (length <= 1)
                return;

            if (kdNode[root].Min > CCL.TresholdValue) // test whether is necessary passing through both subtree
            {
                if (dm < root) WalkThroughMaxLevel(dm, root - dm);
            }
            else
            {
                if (dm < root) WalkThroughMaxLevel(dm, root - dm); // left subtree
                if (hm > root + 1) WalkThroughMaxLevel(root + 1, hm - (root + 1)); // right subtree
            }
        }

        /// <summary>
        /// Passing through the tree at max level and searching active cells.
        /// </summary>
        /// <param name="dm">lower limit</param>
        /// <param name="length">umber cells</param>
        private void WalkThroughMaxLevel(int dm, int length)
        {
            int root = dm + length / 2;
            int hm = dm + length;

            // processing of root
            if ((kdNode[root].Min <= CCL.TresholdValue) && kdNode[root].Max >= CCL.TresholdValue)
                Tetrahedronize(kdCell[kdNode[root].pKdCell]);

            if (length <= 1)
                return;

            if (kdNode[root].Max < CCL.TresholdValue) // test whether is necessary passing through both subtree
            {
                if (hm > root + 1) WalkThroughMinLevel(root + 1, hm - (root + 1)); // right subtree
            }
            else
            {
                if (dm < root) WalkThroughMinLevel(dm, root - dm); // left subtree
                if (hm > root + 1) WalkThroughMinLevel(root + 1, hm - (root + 1)); // right subtree
            }
        }

        /// <summary>
        /// Divide cell into tetrahedrons. Determination edges which are intersect by iso-surface.
        /// </summary>
        /// <param name="kdCell">cell</param>
        private void Tetrahedronize(KdCell kdCell)
        {
            Cell cell = kdCell.cell;

            // vertices coordinates
            Point3D e = new Point3D(kdCell.i, kdCell.j, kdCell.k);
            Point3D f = new Point3D(kdCell.i + this.delta.X, kdCell.j, kdCell.k);
            Point3D g = new Point3D(kdCell.i + this.delta.X, kdCell.j, kdCell.k + this.delta.Z);
            Point3D h = new Point3D(kdCell.i, kdCell.j, kdCell.k + this.delta.Z);

            Point3D a = new Point3D(kdCell.i, kdCell.j + this.delta.Y, kdCell.k);
            Point3D b = new Point3D(kdCell.i + this.delta.X, kdCell.j + this.delta.Y, kdCell.k);
            Point3D c = new Point3D(kdCell.i + this.delta.X, kdCell.j + this.delta.Y, kdCell.k + this.delta.Z);
            Point3D d = new Point3D(kdCell.i, kdCell.j + this.delta.Y, kdCell.k + this.delta.Z);

            Point3D s = e + halfDelta;

            // 6 neighbouring centres 
            Centers adjacentCenters = kdCell.centers;

            // normals
            Normals n = kdCell.normals;

            Tetrahedron tHedron = new Tetrahedron();

            int iE = (2 * kdCell.k * this.slice) + (2 * kdCell.j * this.row) + (2 * kdCell.i); // vrchol E
            int iF = iE + 2; // F
            int iH = iE + 2 * this.slice; // H
            int iG = iH + 2; // G

            int iA = iE + 2 * this.row;
            int iB = iA + 2;
            int iD = iH + 2 * this.row;
            int iC = iD + 2;

            int iS = iE + this.slice + this.row + 1;

            Array.Clear(this.localBuffer, 0, this.localBuffer.Length);

            // front side cell tetrahedrons
            #region Front side

            // front centers coordintes
            Point3D front = new Point3D(s.X - this.delta.X, s.Y, s.Z);
            int iFront = iS - 2;

            // assignment values to tetrahedron
            tHedron.SetValues(adjacentCenters.Front, cell.A, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate()) // if values aren't same and iso-surface intersect some edge
            {
                tHedron.SetIndex(iFront, iA, iE, iS);
                tHedron.SetIndices(Octahedron.topIndices); // set edges from octahedron, octahedron is used as pointer to shared edges
                Intersections(tHedron, front, a, e, s, n.NFront, n.NA, n.NE, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.E, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iE, iH, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, front, e, h, s, n.NFront, n.NE, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.H, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iH, iD, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, front, h, d, s, n.NFront, n.NH, n.ND, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.D, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iD, iA, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, front, d, a, s, n.NFront, n.ND, n.NA, n.NS);
            }

            localBuffer[0] = oHedron.Edges[5]; // copy shared edges from octahedron (defined by 4 tetrahedrons)
            localBuffer[2] = oHedron.Edges[6]; // to local buffer. Local buffer is used for edges shared between
            localBuffer[12] = oHedron.Edges[7]; // octahedrons
            localBuffer[9] = oHedron.Edges[8];
            localBuffer[16] = oHedron.Edges[9];
            localBuffer[19] = oHedron.Edges[10];
            localBuffer[7] = oHedron.Edges[11];
            localBuffer[15] = oHedron.Edges[12];

            #endregion

            this.oHedron.Clear(); // clear shared edges between tetraherons

            #region Left side

            oHedron.Edges[6] = localBuffer[9];
            oHedron.Edges[7] = localBuffer[16];
            oHedron.Edges[12] = localBuffer[19];

            // souradnice vrcholu a tetrahedronu
            Point3D left = new Point3D(s.X, s.Y - this.delta.Y, s.Z);
            int iLeft = iS - 2 * this.row;

            tHedron.SetValues(adjacentCenters.Left, cell.E, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iE, iF, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, left, e, f, s, n.NLeft, n.NE, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.F, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iF, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, left, f, g, s, n.NLeft, n.NF, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.G, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iG, iH, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, left, g, h, s, n.NLeft, n.NG, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.H, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iH, iE, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, left, h, e, s, n.NLeft, n.NH, n.NE, n.NS);
            }

            localBuffer[8] = oHedron.Edges[5];
            localBuffer[10] = oHedron.Edges[8];
            localBuffer[17] = oHedron.Edges[9];
            localBuffer[18] = oHedron.Edges[10];
            localBuffer[11] = oHedron.Edges[11];

            #endregion

            this.oHedron.Clear();

            #region Top side

            oHedron.Edges[12] = localBuffer[16];
            oHedron.Edges[11] = localBuffer[8];
            oHedron.Edges[10] = localBuffer[17];
            oHedron.Edges[7] = localBuffer[12];
            oHedron.Edges[6] = localBuffer[0];

            Point3D top = new Point3D(s.X, s.Y, s.Z - this.delta.Z);
            int iTop = iS - 2 * this.slice;

            tHedron.SetValues(adjacentCenters.Top, cell.A, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iA, iB, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, top, a, b, s, n.NTop, n.NA, n.NB, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.B, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iB, iF, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, top, b, f, s, n.NTop, n.NB, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.F, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iF, iE, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, top, f, e, s, n.NTop, n.NF, n.NE, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.E, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iE, iA, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, top, e, a, s, n.NTop, n.NE, n.NA, n.NS);
            }

            localBuffer[1] = oHedron.Edges[5];
            localBuffer[4] = oHedron.Edges[8];
            localBuffer[13] = oHedron.Edges[9];

            #endregion

            this.oHedron.Clear();

            #region Down side

            oHedron.Edges[10] = localBuffer[18];
            oHedron.Edges[9] = localBuffer[19];
            oHedron.Edges[8] = localBuffer[11];
            oHedron.Edges[7] = localBuffer[15];
            oHedron.Edges[5] = localBuffer[7];

            Point3D down = new Point3D(s.X, s.Y, s.Z + this.delta.Z);
            int iDown = iS + 2 * this.slice;

            tHedron.SetValues(adjacentCenters.Down, cell.D, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iD, iH, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, down, d, h, s, n.NDown, n.ND, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.H, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iH, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, down, h, g, s, n.NDown, n.NH, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.G, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iG, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, down, g, c, s, n.NDown, n.NG, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.C, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iC, iD, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, down, c, d, s, n.NDown, n.NC, n.ND, n.NS);
            }

            localBuffer[5] = oHedron.Edges[6];
            localBuffer[6] = oHedron.Edges[11];
            localBuffer[14] = oHedron.Edges[12];

            #endregion

            this.oHedron.Clear();

            #region Right side

            oHedron.Edges[5] = localBuffer[1];
            oHedron.Edges[7] = localBuffer[13];
            oHedron.Edges[8] = localBuffer[2];
            oHedron.Edges[9] = localBuffer[12];
            oHedron.Edges[10] = localBuffer[15];
            oHedron.Edges[11] = localBuffer[5];
            oHedron.Edges[12] = localBuffer[14];

            Point3D right = new Point3D(s.X, s.Y + this.delta.Y, s.Z);
            int iRight = iS + 2 * this.row;

            tHedron.SetValues(adjacentCenters.Right, cell.B, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iB, iA, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, right, b, a, s, n.NRight, n.NB, n.NA, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.A, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iA, iD, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, right, a, d, s, n.NRight, n.NA, n.ND, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.D, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iD, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, right, d, c, s, n.NRight, n.ND, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.C, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iC, iB, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, right, c, b, s, n.NRight, n.NC, n.NB, n.NS);
            }

            localBuffer[3] = oHedron.Edges[6];

            #endregion

            this.oHedron.Clear();

            #region Back side

            oHedron.Edges[5] = localBuffer[4];
            oHedron.Edges[6] = localBuffer[3];
            oHedron.Edges[7] = localBuffer[13];
            oHedron.Edges[8] = localBuffer[10];
            oHedron.Edges[9] = localBuffer[17];
            oHedron.Edges[10] = localBuffer[18];
            oHedron.Edges[11] = localBuffer[6];
            oHedron.Edges[12] = localBuffer[14];

            Point3D back = new Point3D(s.X + this.delta.X, s.Y, s.Z);
            int iBack = iS + 2;

            tHedron.SetValues(adjacentCenters.Back, cell.B, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iB, iF, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, back, b, f, s, n.NBack, n.NB, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.F, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iF, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, back, f, g, s, n.NBack, n.NF, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.G, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iG, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, back, g, c, s, n.NBack, n.NG, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.C, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iC, iB, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, back, c, b, s, n.NBack, n.NC, n.NB, n.NS);
            }

            #endregion

            this.oHedron.Clear();
        }

        /// <summary>
        /// Decision whether apend cell into kd-tree. If are all values same then cell isn't append.
        /// </summary>
        /// <param name="i">column index</param>
        /// <param name="j">row index</param>
        /// <param name="k">slice index</param>
        private void SplitToCells(int i, int j, int k)
        {
#if(DEBUG)
            if (i > (this.row - 2) && j > (this.row - 2) && k > (this.row - 2))
                throw new IndexOutOfRangeException("Index out of range [Tetrahedronize]");
#endif
            Cell cell;

            if (this.middleBuffer[i + j * this.row].Used)
                cell = this.middleBuffer[i + j * this.row];
            else
            {
                cell = GetCell(i, j, k);
                this.middleBuffer[i + j * this.row] = cell;
            }

            // 6 neigbouring centres
            Centers adjacentCenters;
            GetAdjacentCenters(i, j, k, out adjacentCenters);

            // all values are same
            if (adjacentCenters.Useless() && adjacentCenters.Front == cell.S)
                return;

            // normals
            Normals n = ComputeNormals(i, j, k, adjacentCenters, cell);

            // add cell into array of cells, which are used for kd-tree
            this.kdCell[kdCells++] = new KdCell(cell, adjacentCenters, n, i, j, k);
        }

        /// <summary>
        /// Create kd-tree. First is created array of cells a then is created kd-tree.
        /// </summary>
        private void BuildKdTree()
        {
            long dt = Environment.TickCount;

            this.ProgressInfo(0f);

            // cells for extraction
            for (int k = 1; k < (this.size[2] - 2); k++)
            {
                for (int j = 1; j < this.size[1] - 2; j++)
                {
                    for (int i = 1 + (j % 2 + k % 2); i < this.size[0] - 2; i += 2)
                    {
                        SplitToCells(i, j, k);
                    }
                }

                this.ProgressInfo(((float)k / this.size[2]) * 0.3f);
                this.SwapBuffers();
            }

            // kd-tree nodes array, kd-tree is defined on this array
            this.kdNode = new KdNode[this.kdCells];

            int block = 10000;

            // fill kd-nodes with parametrs of cells
            for (int i = 0; i < this.kdCells; i++)
            {
                this.kdNode[i] = new KdNode(i, this.kdCell[i].GetMin(), this.kdCell[i].GetMax());

                if (i == block)
                {
                    this.ProgressInfo(0.3f + (((float)i / this.kdCells) * 0.2f));
                    block += 10000;
                }
            }

            this.ProgressInfo(0.5f);

            // recursive building of tree
            MinLevel(0, this.kdCells);

            this.ProgressInfo(0.8f);

            Console.WriteLine(" Time of build kd-tree: {0} ms", Environment.TickCount - dt);
        }

        #endregion

        #region Classic CCL

        /// <summary>
        /// Passing through cells.
        /// </summary>
        private void WalkThrough()
        {
            for (int k = 1; k < (this.size[2] - 2); k++)
            {
                for (int j = 1; j < (this.size[1] - 2); j++)
                {
                    for (int i = 1 + (j % 2 + k % 2); i < this.size[0] - 2; i += 2)
                    {
                        Tetrahedronize(i, j, k); // divide cell to 24 tetrahedrons
                    }
                }

                this.SwapBuffers();
                this.ProgressInfo(((float)k / this.size[2]) * 0.8f);
            }
        }

        /// <summary>
        /// Swap buffers.
        /// </summary>
        private void SwapBuffers()
        {
            Cell[] tmp = this.upBuffer;

            this.upBuffer = this.middleBuffer;
            this.middleBuffer = this.downBuffer;
            this.downBuffer = tmp;

            Array.Clear(this.downBuffer, 0, this.downBuffer.Length);

            tmp = null;
        }

        /// <summary>
        /// Divide cell into tetrahedrons. Determination edges which are intersect by iso-surface. 
        /// </summary>
        /// <param name="i">column index</param>
        /// <param name="j">row index</param>
        /// <param name="k">slice index</param>
        private void Tetrahedronize(int i, int j, int k)
        {
#if(DEBUG)
            if (i > (this.row - 2) && j > (this.row - 2) && k > (this.row - 2))
                throw new IndexOutOfRangeException("Index out of range [Tetrahedronize]");
#endif
            Cell cell;

            if (this.middleBuffer[i + j * this.row].Used)
                cell = this.middleBuffer[i + j * this.row];
            else
            {
                cell = GetCell(i, j, k);
                this.middleBuffer[i + j * this.row] = cell;
            }

            // urceni 6 okolnich stredu
            Centers adjacentCenters;
            GetAdjacentCenters(i, j, k, out adjacentCenters);

            // pokud vsechny hodnoty jsou stejne
            if (adjacentCenters.Useless() && adjacentCenters.Front == cell.S)
                return;

            // souradnice vrcholu bunky
            Point3D e = new Point3D(i, j, k);
            Point3D f = new Point3D(i + this.delta.X, j, k);
            Point3D g = new Point3D(i + this.delta.X, j, k + this.delta.Z);
            Point3D h = new Point3D(i, j, k + this.delta.Z);

            Point3D a = new Point3D(i, j + this.delta.Y, k);
            Point3D b = new Point3D(i + this.delta.X, j + this.delta.Y, k);
            Point3D c = new Point3D(i + this.delta.X, j + this.delta.Y, k + this.delta.Z);
            Point3D d = new Point3D(i, j + this.delta.Y, k + this.delta.Z);

            Normals n = ComputeNormals(i, j, k, adjacentCenters, cell);

            Point3D s = e + halfDelta;

            Tetrahedron tHedron = new Tetrahedron();

            Array.Clear(this.localBuffer, 0, this.localBuffer.Length);

            int iE = (2 * k * this.slice) + (2 * j * this.row) + (2 * i); // vrchol E
            int iF = iE + 2; // F
            int iH = iE + 2 * this.slice; // H
            int iG = iH + 2; // G

            int iA = iE + 2 * this.row;
            int iB = iA + 2;
            int iD = iH + 2 * this.row;
            int iC = iD + 2;

            int iS = iE + this.slice + this.row + 1;

            // front side cell tetrahedrons
            #region Front side

            // front centers coordintes
            Point3D front = new Point3D(s.X - this.delta.X, s.Y, s.Z);
            int iFront = iS - 2;

            // assignment values to tetrahedron
            tHedron.SetValues(adjacentCenters.Front, cell.A, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate()) // if values aren't same and iso-surface intersect some edge
            {
                tHedron.SetIndex(iFront, iA, iE, iS);
                tHedron.SetIndices(Octahedron.topIndices); // set edges from octahedron, octahedron is used as pointer to shared edges
                Intersections(tHedron, front, a, e, s, n.NFront, n.NA, n.NE, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.E, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iE, iH, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, front, e, h, s, n.NFront, n.NE, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.H, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iH, iD, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, front, h, d, s, n.NFront, n.NH, n.ND, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Front, cell.D, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iFront, iD, iA, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, front, d, a, s, n.NFront, n.ND, n.NA, n.NS);
            }

            localBuffer[0] = oHedron.Edges[5]; // copy shared edges from octahedron (defined by 4 tetrahedrons)
            localBuffer[2] = oHedron.Edges[6]; // to local buffer. Local buffer is used for edges shared between
            localBuffer[12] = oHedron.Edges[7]; // octahedrons
            localBuffer[9] = oHedron.Edges[8];
            localBuffer[16] = oHedron.Edges[9];
            localBuffer[19] = oHedron.Edges[10];
            localBuffer[7] = oHedron.Edges[11];
            localBuffer[15] = oHedron.Edges[12];

            #endregion

            this.oHedron.Clear(); // clear shared edges between tetraherons

            #region Left side

            oHedron.Edges[6] = localBuffer[9];
            oHedron.Edges[7] = localBuffer[16];
            oHedron.Edges[12] = localBuffer[19];

            // souradnice vrcholu a tetrahedronu
            Point3D left = new Point3D(s.X, s.Y - this.delta.Y, s.Z);
            int iLeft = iS - 2 * this.row;

            tHedron.SetValues(adjacentCenters.Left, cell.E, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iE, iF, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, left, e, f, s, n.NLeft, n.NE, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.F, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iF, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, left, f, g, s, n.NLeft, n.NF, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.G, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iG, iH, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, left, g, h, s, n.NLeft, n.NG, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Left, cell.H, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iLeft, iH, iE, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, left, h, e, s, n.NLeft, n.NH, n.NE, n.NS);
            }

            localBuffer[8] = oHedron.Edges[5];
            localBuffer[10] = oHedron.Edges[8];
            localBuffer[17] = oHedron.Edges[9];
            localBuffer[18] = oHedron.Edges[10];
            localBuffer[11] = oHedron.Edges[11];

            #endregion

            this.oHedron.Clear();

            #region Top side

            oHedron.Edges[12] = localBuffer[16];
            oHedron.Edges[11] = localBuffer[8];
            oHedron.Edges[10] = localBuffer[17];
            oHedron.Edges[7] = localBuffer[12];
            oHedron.Edges[6] = localBuffer[0];

            Point3D top = new Point3D(s.X, s.Y, s.Z - this.delta.Z);
            int iTop = iS - 2 * this.slice;

            tHedron.SetValues(adjacentCenters.Top, cell.A, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iA, iB, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, top, a, b, s, n.NTop, n.NA, n.NB, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.B, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iB, iF, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, top, b, f, s, n.NTop, n.NB, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.F, cell.E, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iF, iE, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, top, f, e, s, n.NTop, n.NF, n.NE, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Top, cell.E, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iTop, iE, iA, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, top, e, a, s, n.NTop, n.NE, n.NA, n.NS);
            }

            localBuffer[1] = oHedron.Edges[5];
            localBuffer[4] = oHedron.Edges[8];
            localBuffer[13] = oHedron.Edges[9];

            #endregion

            this.oHedron.Clear();

            #region Down side

            oHedron.Edges[10] = localBuffer[18];
            oHedron.Edges[9] = localBuffer[19];
            oHedron.Edges[8] = localBuffer[11];
            oHedron.Edges[7] = localBuffer[15];
            oHedron.Edges[5] = localBuffer[7];

            Point3D down = new Point3D(s.X, s.Y, s.Z + this.delta.Z);
            int iDown = iS + 2 * this.slice;

            tHedron.SetValues(adjacentCenters.Down, cell.D, cell.H, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iD, iH, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, down, d, h, s, n.NDown, n.ND, n.NH, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.H, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iH, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, down, h, g, s, n.NDown, n.NH, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.G, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iG, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, down, g, c, s, n.NDown, n.NG, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Down, cell.C, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iDown, iC, iD, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, down, c, d, s, n.NDown, n.NC, n.ND, n.NS);
            }

            localBuffer[5] = oHedron.Edges[6];
            localBuffer[6] = oHedron.Edges[11];
            localBuffer[14] = oHedron.Edges[12];

            #endregion

            this.oHedron.Clear();

            #region Right side

            oHedron.Edges[5] = localBuffer[1];
            oHedron.Edges[7] = localBuffer[13];
            oHedron.Edges[8] = localBuffer[2];
            oHedron.Edges[9] = localBuffer[12];
            oHedron.Edges[10] = localBuffer[15];
            oHedron.Edges[11] = localBuffer[5];
            oHedron.Edges[12] = localBuffer[14];

            Point3D right = new Point3D(s.X, s.Y + this.delta.Y, s.Z);
            int iRight = iS + 2 * this.row;

            tHedron.SetValues(adjacentCenters.Right, cell.B, cell.A, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iB, iA, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, right, b, a, s, n.NRight, n.NB, n.NA, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.A, cell.D, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iA, iD, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, right, a, d, s, n.NRight, n.NA, n.ND, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.D, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iD, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, right, d, c, s, n.NRight, n.ND, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Right, cell.C, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iRight, iC, iB, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, right, c, b, s, n.NRight, n.NC, n.NB, n.NS);
            }

            localBuffer[3] = oHedron.Edges[6];

            #endregion

            this.oHedron.Clear();

            #region Back side

            oHedron.Edges[5] = localBuffer[4];
            oHedron.Edges[6] = localBuffer[3];
            oHedron.Edges[7] = localBuffer[13];
            oHedron.Edges[8] = localBuffer[10];
            oHedron.Edges[9] = localBuffer[17];
            oHedron.Edges[10] = localBuffer[18];
            oHedron.Edges[11] = localBuffer[6];
            oHedron.Edges[12] = localBuffer[14];

            Point3D back = new Point3D(s.X + this.delta.X, s.Y, s.Z);
            int iBack = iS + 2;

            tHedron.SetValues(adjacentCenters.Back, cell.B, cell.F, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iB, iF, iS);
                tHedron.SetIndices(Octahedron.topIndices);
                Intersections(tHedron, back, b, f, s, n.NBack, n.NB, n.NF, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.F, cell.G, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iF, iG, iS);
                tHedron.SetIndices(Octahedron.backIndices);
                Intersections(tHedron, back, f, g, s, n.NBack, n.NF, n.NG, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.G, cell.C, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iG, iC, iS);
                tHedron.SetIndices(Octahedron.downIndices);
                Intersections(tHedron, back, g, c, s, n.NBack, n.NG, n.NC, n.NS);
            }

            tHedron.SetValues(adjacentCenters.Back, cell.C, cell.B, cell.S);
            if (!tHedron.Useless && tHedron.Evaluate())
            {
                tHedron.SetIndex(iBack, iC, iB, iS);
                tHedron.SetIndices(Octahedron.frontIndices);
                Intersections(tHedron, back, c, b, s, n.NBack, n.NC, n.NB, n.NS);
            }

            #endregion

            this.oHedron.Clear();
        }

        /// <summary>
        /// Determination intersetion of tetrahedron.
        /// </summary>
        private void Intersections(Tetrahedron hedron, Point3D a, Point3D b, Point3D c, Point3D d, Normal3D na, Normal3D nb, Normal3D nc, Normal3D nd)
        {
            DEdge ac, bc, ab, bd, ad, cd;
            int indicia, index0, index1;

            switch (hedron.Flags.Value)
            {
                case 1: // d
                case 14:
                    if (this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB);
                    else
                    {
                        bd = new DEdge(hedron.ib, hedron.id);

                        if (this.dictionary.TryGetValue(bd, out indicia)) // slice
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection = true;
                            Intersection(b, d, nb, nd, hedron.B, hedron.D);
                            this.dictionary.Add(bd, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB);
                    else
                    {
                        ad = new DEdge(hedron.ia, hedron.id);

                        if (this.dictionary.TryGetValue(ad, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection = true;
                            Intersection(a, d, na, nd, hedron.A, hedron.D);
                            this.dictionary.Add(ad, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB);
                    else
                    {
                        cd = new DEdge(hedron.ic, hedron.id);

                        if (this.dictionary.TryGetValue(cd, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection = true;
                            Intersection(c, d, nc, nd, hedron.C, hedron.D);
                            this.dictionary.Add(cd, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    break;

                case 2: // c
                case 13:
                    if (this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB);
                    else
                    {
                        ac = new DEdge(hedron.ia, hedron.ic);

                        if (this.dictionary.TryGetValue(ac, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection = true;
                            Intersection(a, c, na, nc, hedron.A, hedron.C);
                            this.dictionary.Add(ac, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection)
                    {
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB);
                    }
                    else
                    {
                        bc = new DEdge(hedron.ib, hedron.ic);

                        if (this.dictionary.TryGetValue(bc, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection = true;
                            Intersection(b, c, nb, nc, hedron.B, hedron.C);
                            this.dictionary.Add(bc, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB);
                    else
                    {
                        cd = new DEdge(hedron.ic, hedron.id);

                        if (this.dictionary.TryGetValue(cd, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection = true;
                            Intersection(c, d, nc, nd, hedron.C, hedron.D);
                            this.dictionary.Add(cd, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    break;

                case 4: // b
                case 11:
                    if (this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB);
                    else
                    {
                        ab = new DEdge(hedron.ia, hedron.ib);
                        if (this.dictionary.TryGetValue(ab, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection = true;
                            Intersection(a, b, na, nb, hedron.A, hedron.B);
                            this.dictionary.Add(ab, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB);
                    else
                    {
                        bc = new DEdge(hedron.ib, hedron.ic);

                        if (this.dictionary.TryGetValue(bc, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection = true;
                            Intersection(b, c, nb, nc, hedron.B, hedron.C);
                            this.dictionary.Add(bc, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection) // local
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB);
                    else
                    {
                        bd = new DEdge(hedron.ib, hedron.id);

                        if (this.dictionary.TryGetValue(bd, out indicia)) // slice
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection = true;
                            Intersection(b, d, nb, nd, hedron.B, hedron.D);
                            this.dictionary.Add(bd, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }
                    break;

                case 7: // a
                case 8:
                    if (this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB);
                    else
                    {
                        ab = new DEdge(hedron.ia, hedron.ib);
                        if (this.dictionary.TryGetValue(ab, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection = true;
                            Intersection(a, b, na, nb, hedron.A, hedron.B);
                            this.dictionary.Add(ab, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB);
                    else
                    {
                        ac = new DEdge(hedron.ia, hedron.ic);

                        if (this.dictionary.TryGetValue(ac, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection = true;
                            Intersection(a, c, na, nc, hedron.A, hedron.C);
                            this.dictionary.Add(ac, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB);
                    else
                    {
                        ad = new DEdge(hedron.ia, hedron.id);

                        if (this.dictionary.TryGetValue(ad, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection = true;
                            Intersection(a, d, na, nd, hedron.A, hedron.D);
                            this.dictionary.Add(ad, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }
                    break;

                case 3: // AC AD, BC, BD
                case 12:
                    if (this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB);
                    else
                    {
                        ac = new DEdge(hedron.ia, hedron.ic);

                        if (this.dictionary.TryGetValue(ac, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection = true;
                            Intersection(a, c, na, nc, hedron.A, hedron.C);
                            this.dictionary.Add(ac, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection)
                    {
                        index0 = this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB;
                        this.indices.Add(index0);
                    }
                    else
                    {
                        ad = new DEdge(hedron.ia, hedron.id);

                        if (this.dictionary.TryGetValue(ad, out indicia))
                        {
                            index0 = indicia;
                            this.indices.Add(index0);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection = true;
                            Intersection(a, d, na, nd, hedron.A, hedron.D);
                            index0 = indicesValue;
                            this.dictionary.Add(ad, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection)
                    {
                        index1 = this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB;
                        this.indices.Add(index1);
                    }
                    else
                    {
                        bc = new DEdge(hedron.ib, hedron.ic);

                        if (this.dictionary.TryGetValue(bc, out indicia))
                        {
                            index1 = indicia;
                            this.indices.Add(index1);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection = true;
                            Intersection(b, c, nb, nc, hedron.B, hedron.C);
                            index1 = indicesValue;
                            this.dictionary.Add(bc, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection)
                    {
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB);
                        this.indices.Add(index1);
                        this.indices.Add(index0);
                    }
                    else
                    {
                        bd = new DEdge(hedron.ib, hedron.id);

                        if (this.dictionary.TryGetValue(bd, out indicia))
                        {
                            this.indices.Add(this.dictionary[bd]);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection = true;
                            Intersection(b, d, nb, nd, hedron.B, hedron.D);
                            this.dictionary.Add(bd, this.indicesValue);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                            this.indices.Add(indicesValue++);
                        }
                    }
                    break;

                case 5: // AB, AD, BC, CD
                case 10:
                    if (this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB);
                    else
                    {
                        ab = new DEdge(hedron.ia, hedron.ib);
                        if (this.dictionary.TryGetValue(ab, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection = true;
                            Intersection(a, b, na, nb, hedron.A, hedron.B);
                            this.dictionary.Add(ab, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection)
                    {
                        index0 = this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB;
                        this.indices.Add(index0);
                    }
                    else
                    {
                        ad = new DEdge(hedron.ia, hedron.id);

                        if (this.dictionary.TryGetValue(ad, out indicia))
                        {
                            index0 = indicia;
                            this.indices.Add(index0);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[2]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[2]].HasIntersection = true;
                            index0 = indicesValue;
                            Intersection(a, d, na, nd, hedron.A, hedron.D);
                            this.dictionary.Add(ad, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection)
                    {
                        index1 = this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB;
                        this.indices.Add(index1);
                    }
                    else
                    {
                        bc = new DEdge(hedron.ib, hedron.ic);

                        if (this.dictionary.TryGetValue(bc, out indicia))
                        {
                            index1 = indicia;
                            this.indices.Add(index1);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[3]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[3]].HasIntersection = true;
                            index1 = indicesValue;
                            Intersection(b, c, nb, nc, hedron.B, hedron.C);
                            this.dictionary.Add(bc, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection)
                    {
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB);
                        this.indices.Add(index1);
                        this.indices.Add(index0);
                    }
                    else
                    {
                        cd = new DEdge(hedron.ic, hedron.id);

                        if (this.dictionary.TryGetValue(cd, out indicia))
                        {
                            this.indices.Add(this.dictionary[cd]);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection = true;
                            Intersection(c, d, nc, nd, hedron.C, hedron.D);
                            this.dictionary.Add(cd, indicesValue);
                            this.indices.Add(indicesValue++);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                        }
                    }
                    break;

                case 6: // AB, AC, BD, CD
                case 9:
                    if (this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection)
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB);
                    else
                    {
                        ab = new DEdge(hedron.ia, hedron.ib);
                        if (this.dictionary.TryGetValue(ab, out indicia))
                            this.indices.Add(indicia);
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[0]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[0]].HasIntersection = true;
                            Intersection(a, b, na, nb, hedron.A, hedron.B);
                            this.dictionary.Add(ab, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection)
                    {
                        index0 = this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB;
                        this.indices.Add(index0);
                    }
                    else
                    {
                        ac = new DEdge(hedron.ia, hedron.ic);

                        if (this.dictionary.TryGetValue(ac, out indicia))
                        {
                            index0 = indicia;
                            this.indices.Add(indicia);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[1]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[1]].HasIntersection = true;
                            Intersection(a, c, na, nc, hedron.A, hedron.C);
                            index0 = this.indicesValue;
                            this.dictionary.Add(ac, this.indicesValue);
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection) // local
                    {
                        index1 = this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB;
                        this.indices.Add(index1);
                    }
                    else
                    {
                        bd = new DEdge(hedron.ib, hedron.id);

                        if (this.dictionary.TryGetValue(bd, out indicia)) // slice
                        {
                            this.indices.Add(indicia);
                            index1 = indicia;
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[4]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[4]].HasIntersection = true;
                            Intersection(b, d, nb, nd, hedron.B, hedron.D);
                            this.dictionary.Add(bd, this.indicesValue);
                            index1 = indicesValue;
                            this.indices.Add(indicesValue++);
                        }
                    }

                    if (this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection)
                    {
                        this.indices.Add(this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB);
                        this.indices.Add(index1);
                        this.indices.Add(index0);
                    }
                    else
                    {
                        cd = new DEdge(hedron.ic, hedron.id);

                        if (this.dictionary.TryGetValue(cd, out indicia))
                        {
                            this.indices.Add(this.dictionary[cd]);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                        }
                        else
                        {
                            this.oHedron.Edges[hedron.EdgeIndices[5]].IndexIB = indicesValue;
                            this.oHedron.Edges[hedron.EdgeIndices[5]].HasIntersection = true;
                            Intersection(c, d, nc, nd, hedron.C, hedron.D);
                            this.dictionary.Add(cd, indicesValue);
                            this.indices.Add(indicesValue++);
                            this.indices.Add(index1);
                            this.indices.Add(index0);
                        }
                    }
                    break;
            }
        }

        /// <summary>
        /// Edge intersection a determiantion normal in intersection.
        /// </summary>
        /// <param name="a">coordinates vertices a</param>
        /// <param name="b">coordinates vertices b</param>
        /// <param name="na">normal coordinates in vertices a</param>
        /// <param name="nb">normal coordinates in vertices b</param>
        /// <param name="va">values in vertices a</param>
        /// <param name="vb">values in vertices b</param>
        public void Intersection(Point3D a, Point3D b, Normal3D na, Normal3D nb, double va, double vb)
        {
            double tmp = (CCL.TresholdValue - va) / (vb - va);

            Point3D p = a + (b - a) * tmp;
            Normal3D n = na + (nb - na) * tmp;

            p.X *= this.distortion.X;
            p.Y *= this.distortion.Y;
            p.Z *= this.distortion.Z;

            n.Nx *= this.distortion.X;
            n.Ny *= this.distortion.Y;
            n.Nz *= this.distortion.Z;

            this.vertices.Add(p);
            this.normals.Add(n);
        }

        /// <summary>
        /// Determination values of six neighbouring centres.
        /// </summary>
        private void GetAdjacentCenters(int i, int j, int k, out Centers adjacentCenters)
        {
#if(DEBUG)
            if ((i < 1 && j < 1 && k < 1) || (i > (this.row - 2) && j > (this.row - 2) && k > (this.row - 2)))
                throw new IndexOutOfRangeException("Index out of range [GetAdjecentCentres]");
#endif
            Cell cell;

            adjacentCenters = new Centers();

            long index = i + j * this.row;

            // front
            if (this.middleBuffer[index - 1].Used)
                adjacentCenters.Front = this.middleBuffer[index - 1].S;
            else
            {
                cell = GetCell(i - 1, j, k);
                this.middleBuffer[index - 1] = cell;
                adjacentCenters.Front = cell.S;
            }

            // back
            if (this.middleBuffer[index + 1].Used)
                adjacentCenters.Back = this.middleBuffer[index + 1].S;
            else
            {
                cell = GetCell(i + 1, j, k);
                this.middleBuffer[index + 1] = cell;
                adjacentCenters.Back = cell.S;
            }

            // left
            if (this.middleBuffer[index - this.row].Used)
                adjacentCenters.Left = this.middleBuffer[index - this.row].S;
            else
            {
                cell = GetCell(i, j - 1, k);
                this.middleBuffer[index - this.row] = cell;
                adjacentCenters.Left = cell.S;
            }

            // right
            if (this.middleBuffer[index + this.row].Used)
                adjacentCenters.Right = this.middleBuffer[index + this.row].S;
            else
            {
                cell = GetCell(i, j + 1, k);
                this.middleBuffer[index + this.row] = cell;
                adjacentCenters.Right = cell.S;
            }

            // down
            if (this.downBuffer[index].Used)
                adjacentCenters.Down = this.downBuffer[index].S;
            else
            {
                cell = GetCell(i, j, k + 1);
                this.downBuffer[index] = cell;
                adjacentCenters.Down = cell.S;
            }

            // up
            if (this.upBuffer[index].Used)
                adjacentCenters.Top = this.upBuffer[index].S;
            else
            {
                cell = GetCell(i, j, k - 1);
                this.upBuffer[index] = cell;
                adjacentCenters.Top = cell.S;
            }
        }

        /// <summary>
        /// Determination normals.
        /// </summary>
        private Normals ComputeNormals(int i, int j, int k, Centers adjacentCenters, Cell cx)
        {
            Cell front, back, left, right, top, down;
            Cell a, b, c, d, e, f, g, h;
            Normals normals = new Normals();

            #region Cells necessary for normals determination

            int iPlus = i + 1;
            int iMinus = i - 1;
            int jPlus = j + 1;
            int jMinus = j - 1;
            int iPlusPlus = iPlus + 1;
            int iMinusMinus = iMinus - 1;
            int jPlusPlus = jPlus + 1;
            int jMinusMinus = jMinus - 1;

            int jRow = j * this.row;


            // front
            if ((iMinus) > 0)
            {
                if (this.middleBuffer[iMinusMinus + jRow].Used)
                    front = this.middleBuffer[iMinusMinus + jRow];
                else
                {
                    front = GetCell(iMinusMinus, j, k);
                    this.middleBuffer[iMinusMinus + jRow] = front;
                }
            }
            else
                front = this.middleBuffer[iMinus + jRow];

            // back
            if ((iPlus) < (this.size[0] - 2))
            {
                if (this.middleBuffer[iPlusPlus + jRow].Used)
                    back = this.middleBuffer[iPlusPlus + jRow];
                else
                {
                    back = GetCell(iPlusPlus, j, k);
                    this.middleBuffer[iPlusPlus + jRow] = back;
                }
            }
            else
                back = this.middleBuffer[iPlus + jRow];

            // left
            if (jMinus > 0)
            {
                if (this.middleBuffer[i + jMinusMinus * this.row].Used)
                    left = this.middleBuffer[i + jMinusMinus * this.row];
                else
                {
                    left = GetCell(i, jMinusMinus, k);
                    this.middleBuffer[i + jMinusMinus * this.row] = left;
                }
            }
            else
                left = this.middleBuffer[i + jMinus * this.row];

            // right
            if (jPlus < (this.size[1] - 2))
            {
                if (this.middleBuffer[i + jPlusPlus * this.row].Used)
                    right = this.middleBuffer[i + jPlusPlus * this.row];
                else
                {
                    right = GetCell(i, jPlusPlus, k);
                    this.middleBuffer[i + jPlusPlus * this.row] = right;
                }
            }
            else
                right = this.middleBuffer[i + jPlus * this.row];

            // top
            if ((k - 1) > 0)
                top = GetCell(i, j, k - 2);
            else
                top = this.upBuffer[i + jRow];

            // down
            if ((k + 1) < (this.size[2] - 2))
                down = GetCell(i, j, k + 2);
            else
                down = this.upBuffer[i + jRow];

            // e
            if (this.upBuffer[iMinus + jMinus * this.row].Used)
                e = this.upBuffer[iMinus + jMinus * this.row];
            else
            {
                e = GetCell(iMinus, jMinus, k - 1);
                this.upBuffer[iMinus + jMinus * this.row] = e;
            }

            // f
            if (this.upBuffer[iPlus + jMinus * this.row].Used)
                f = this.upBuffer[iPlus + jMinus * this.row];
            else
            {
                f = GetCell(iPlus, jMinus, k - 1);
                this.upBuffer[(iPlus) + jMinus * this.row] = f;
            }

            // g
            if (this.downBuffer[iPlus + jMinus * this.row].Used)
                g = this.downBuffer[iPlus + jMinus * this.row];
            else
            {
                g = GetCell(iPlus, jMinus, k + 1);
                this.downBuffer[iPlus + jMinus * this.row] = g;
            }

            // h
            if (this.downBuffer[iMinus + jMinus * this.row].Used)
                h = this.downBuffer[iMinus + jMinus * this.row];
            else
            {
                h = GetCell(iMinus, jMinus, k + 1);
                this.downBuffer[(iMinus) + jMinus * this.row] = h;
            }

            // a
            if (this.upBuffer[(iMinus) + jPlus * this.row].Used)
                a = this.upBuffer[(iMinus) + jPlus * this.row];
            else
            {
                a = GetCell(iMinus, jPlus, k - 1);
                this.upBuffer[iMinus + jPlus * this.row] = a;
            }

            // b
            if (this.upBuffer[iPlus + jPlus * this.row].Used)
                b = this.upBuffer[iPlus + jPlus * this.row];
            else
            {
                b = GetCell(iPlus, jPlus, k - 1);
                this.upBuffer[iPlus + jPlus * this.row] = b;
            }

            // c
            if (this.downBuffer[(iPlus) + jPlus * this.row].Used)
                c = this.downBuffer[(iPlus) + jPlus * this.row];
            else
            {
                c = GetCell(iPlus, jPlus, k + 1);
                this.downBuffer[iPlus + jPlus * this.row] = c;
            }

            // d
            if (this.downBuffer[iMinus + jPlus * this.row].Used)
                d = this.downBuffer[iMinus + jPlus * this.row];
            else
            {
                d = GetCell(iMinus, jPlus, k + 1);
                this.downBuffer[iMinus + jPlus * this.row] = d;
            }

            #endregion

            #region Normlas in vertices

            normals.NA = new Normal3D(cx.B - front.B, right.E - cx.E, cx.D - top.D);
            normals.NA *= 0.5f;

            normals.NS = normals.NA;

            normals.NB = new Normal3D(back.A - cx.A, right.F - cx.F, cx.C - top.C);
            normals.NB *= 0.5f;
            normals.NS += normals.NB;

            normals.NC = new Normal3D(back.D - cx.D, right.G - cx.G, down.B - cx.B);
            normals.NC *= 0.5f;
            normals.NS += normals.NC;

            normals.ND = new Normal3D(cx.C - front.C, right.H - cx.H, down.A - cx.A);
            normals.ND *= 0.5f;
            normals.NS += normals.ND;

            normals.NE = new Normal3D(cx.F - front.F, cx.A - left.A, cx.H - top.H);
            normals.NE *= 0.5f;
            normals.NS += normals.NE;

            normals.NF = new Normal3D(back.E - cx.E, cx.B - left.B, cx.G - top.G);
            normals.NF *= 0.5f;
            normals.NS += normals.NF;

            normals.NG = new Normal3D(back.H - cx.H, cx.C - left.C, down.F - cx.F);
            normals.NG *= 0.5f;
            normals.NS += normals.NG;

            normals.NH = new Normal3D(cx.G - front.G, cx.D - left.D, down.E - cx.E);
            normals.NH *= 0.5f;
            normals.NS += normals.NH;

            normals.NS *= 0.125f;

            #endregion

            #region Normal neighbourin centers

            // front center normal
            Normal3D x0 = new Normal3D(cx.A - front.A, a.D - front.F, front.C - a.E);
            Normal3D x1 = new Normal3D(cx.E - front.E, front.B - e.H, front.G - e.A);
            Normal3D x2 = new Normal3D(cx.H - front.H, front.C - h.E, h.D - front.F);
            Normal3D x3 = new Normal3D(cx.D - front.D, d.A - front.G, d.H - front.B);

            Normal3D x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NFront = (x + normals.NA + normals.NE + normals.NH + normals.ND) * 0.125f;

            // back center normal
            x0 = new Normal3D(back.B - cx.B, b.C - back.E, back.D - b.F);
            x1 = new Normal3D(back.F - cx.F, back.A - f.G, back.H - f.B);
            x2 = new Normal3D(back.G - cx.G, back.D - g.F, g.C - back.E);
            x3 = new Normal3D(back.C - cx.C, c.B - back.H, c.G - back.A);

            x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NBack = (x + normals.NB + normals.NF + normals.NG + normals.NC) * 0.125f;

            // left center normal
            x0 = new Normal3D(left.B - e.H, cx.E - left.E, left.D - e.F);
            x1 = new Normal3D(f.G - left.A, cx.F - left.F, left.C - f.E);
            x2 = new Normal3D(g.F - left.D, cx.G - left.G, g.H - left.B);
            x3 = new Normal3D(left.C - h.E, cx.H - left.H, h.G - left.A);

            x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NLeft = (x + normals.NE + normals.NF + normals.NG + normals.NH) * 0.125f;

            // right center normal
            x0 = new Normal3D(b.C - right.E, right.B - cx.B, right.G - b.A);
            x1 = new Normal3D(right.F - a.D, right.A - cx.A, right.H - a.B);
            x2 = new Normal3D(right.G - d.A, right.D - cx.D, d.C - right.E);
            x3 = new Normal3D(c.B - right.H, right.C - cx.C, c.D - right.F);

            x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NRight = (x + normals.NA + normals.NB + normals.NC + normals.ND) * 0.125f;

            // top center normal
            x0 = new Normal3D(top.G - e.A, top.D - e.F, cx.E - top.E);
            x1 = new Normal3D(f.B - top.H, top.C - f.E, cx.F - top.F);
            x2 = new Normal3D(b.F - top.D, b.A - top.G, cx.B - top.B);
            x3 = new Normal3D(top.C - a.E, a.B - top.H, cx.A - top.A);

            x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NTop = (x + normals.NA + normals.NB + normals.NE + normals.NF) * 0.125f;

            // down center normal
            x0 = new Normal3D(down.F - h.D, down.A - h.G, down.H - cx.H);
            x1 = new Normal3D(g.C - down.E, down.B - g.H, down.G - cx.G);
            x2 = new Normal3D(c.G - down.A, c.D - down.F, down.C - cx.C);
            x3 = new Normal3D(down.B - d.H, d.C - down.E, down.D - cx.D);

            x = (x0 + x1 + x2 + x3) * 0.5f;
            normals.NDown = (x + normals.NC + normals.ND + normals.NG + normals.NH) * 0.125f;

            #endregion

            return normals;
        }

        /// <summary>
        /// Determination values of the cell.
        /// </summary>
        /// <param name="i">column index</param>
        /// <param name="j">row index</param>
        /// <param name="k">slice index</param>
        /// <returns>cell at position [i, j, k]</returns>
        private Cell GetCell(int i, int j, int k)
        {
            Cell cell = new Cell();
            cell.Used = true;

            cell.SetValues(
                            (Scalar)this.vData[i, j + 1, k],            // A
                            (Scalar)this.vData[i + 1, j + 1, k],        // B
                            (Scalar)this.vData[i + 1, j + 1, k + 1],    // C
                            (Scalar)this.vData[i, j + 1, k + 1],        // D
                            (Scalar)this.vData[i, j, k],                // E
                            (Scalar)this.vData[i + 1, j, k],            // F
                            (Scalar)this.vData[i + 1, j, k + 1],        // G
                            (Scalar)this.vData[i, j, k + 1]             // H
                        );

            return cell;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Define treshold value in %. Value for which wil iso-surface extract.
        /// </summary> 
        [Description("Treshold value. Range is 1% - 100%. \nIf 'useThresholdAsIsoValue' flag is set to True,"
        + " is this value directly used as iso value to be searched (used for implicit surfaces visualization).")]
        public double Treshold
        {
            set
            {
                if (value >= 0 && value <= 100)
                {
                    this.tresholdPercent = value;
                    CCL.TresholdValue = this.min + (this.max - this.min) * this.tresholdPercent * 0.01f;
                    this.paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Values should be in range 1% - 100% !");
                }
            }

            get { return this.tresholdPercent; }
        }

        /// <summary>
        /// Enable kd-tree preprocessing.
        /// </summary>
        [Description("Enable using kd-tree. Kd-tree is used for searching active cells, but is most memory expensive !!")]
        public bool UseKd_tree
        {
            set
            {
                this.kdTree = value;
                this.kdBuild = false;
                this.paramsChanged = true;
            }

            get { return this.kdTree; }
        }

        /// <summary>
        /// Set name of RegGridND point attribute. Default set to Values.
        /// </summary>
        [Description("Set name of RegGridND point attribute. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND normal attribute. Default set to Values.
        /// </summary>
        [Description("Set name of normals point attribute. Default set to \"Normal3D\"")]
        public string NormalAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.normalAttributeName = value;
                    this.paramsChanged = true;
                }
            }
            get { return this.normalAttributeName; }
        }

        /// <summary>
        /// Property for linearInterpolation variable.
        /// </summary>
        [Description("Specifies whether to use input threshold as percentual value from min/max interval or directly as iso value.")]
        public bool UseThresholdAsIsoValue
        {
            get { return (this.useThresholdAsIsoValue); }
            set
            {
                this.useThresholdAsIsoValue = value;
                this.paramsChanged = true;
            }
        }

        #endregion

        #region Internal structures

        /// <summary>
        /// Structure used for keep vol. data which define cell. Values A ... H are read from vol. data source and
        /// sample S is calculated: S = A + B + ... + H /8 (aritmetic mean).
        /// </summary>
        /// <remarks>
        ///   E ------ F
        ///  /|       /|
        /// A ------ B |
        /// | |      | |
        /// | H -----| G
        /// |/       |/
        /// D ------ C
        /// 
        /// </remarks>
        private struct Cell
        {
            /// <summary>
            /// Define whether cell was used. Atribut is used in buffers.
            /// </summary>
            public bool Used;

            /// <summary>
            /// Values in corners.
            /// </summary>
            public double A, B, C, D, E, F, G, H, S;

            /// <summary>
            /// Assingning values.
            /// </summary>
            public void SetValues(double a, double b, double c, double d, double e, double f, double g, double h)
            {
                this.Used = true;

                this.S = this.A = a;
                this.S += this.B = b;
                this.S += this.C = c;
                this.S += this.D = d;
                this.S += this.E = e;
                this.S += this.F = f;
                this.S += this.G = g;
                this.S += this.H = h;

                this.S /= 8;
            }
        }

        /// <summary>
        /// Sturcture used for storage six vaules of surrounding centres.
        /// </summary>
        /// <remarks>
        /// 	      Top
        ///            |  Left          
        ///	           |/                 
        ///  Front ----|---- Back        +--- x
        ///           /|                /|
        ///     Right  |               y |
        ///          Down                z
        /// </remarks>                   
        private struct Centers
        {
            /// <summary>
            /// Centres vaules of surrounding cells.
            /// </summary>
            public double Top, Down, Front, Back, Right, Left;

            /// <summary>
            /// Test whether values are same.
            /// </summary>
            /// <returns>true if values are same</returns>
            public bool Useless()
            {
                if ((Top == Down) && (Down == Front) && (Front == Back) && (Back == Right) && (Right == Left))
                    return true;
                else
                    return false;
            }
        }



        /// <summary>
        /// Structure used for storage information about edge.
        /// </summary>
        private struct Edge
        {
            /// <summary>
            /// End vertices.
            /// </summary>
            public Point3D A, B;
            /// <summary>
            /// Define whether intersection has alredy count.
            /// </summary>
            public bool HasIntersection;

            /// <summary>
            /// Index of intersection. Is valid only if HasIntersetion == true.
            /// </summary>
            public int IndexIB;

            public Edge(Point3D a, Point3D b)
            {
                if (a.X <= b.X && a.Y <= b.Y && a.Z <= b.Z)
                {
                    this.A = a;
                    this.B = b;
                }
                else
                {
                    this.A = b;
                    this.B = a;
                }

                this.HasIntersection = false;
                this.IndexIB = 0;
            }
        }

        /// <summary>
        /// Structure used for storage normals necessary for determination normals in cell.
        /// </summary>
        private struct Normals
        {
            /// <summary>
            /// Normals in vertices of cell and centres.
            /// </summary>
            public Normal3D NA, NB, NC, ND, NE, NF, NG, NH, NS, NFront, NBack, NLeft, NRight, NTop, NDown;
        }

        /// <summary>
        /// Class define 4 tetrahedron where each 2 neighbouring tetrahedrons share 3 edges.
        /// </summary>
        private class Octahedron
        {
            //         0   1   2   3   4   5   6  7   8   9   10  11  12
            // edges: AB, AC, AD, AE, AF, BC, BE, BF, CD, CF, DF, DE, EF;

            /// <summary>
            /// Octahedron edges.
            /// </summary>
            public Edge[] Edges;

            //oHedron.AB, oHedron.AC, oHedron.AF, oHedron.BC, oHedron.BF, oHedron.CF
            public static byte [] topIndices = { 0, 1, 4, 5, 7, 9 };

            //oHedron.AC, oHedron.AD, oHedron.AF, oHedron.CD, oHedron.CF, oHedron.DF
            public static byte [] backIndices = { 1, 2, 4, 8, 9, 10 };

            //oHedron.AD, oHedron.AE, oHedron.AF, oHedron.DE, oHedron.DF, oHedron.EF
            public static byte [] downIndices = { 2, 3, 4, 11, 10, 12 };

            //oHedron.AE, oHedron.AB, oHedron.AF, oHedron.BE, oHedron.EF, oHedron.BF
            public static byte [] frontIndices = { 3, 0, 4, 6, 12, 7 };

            /// <summary>
            /// Create instance.
            /// </summary>
            public Octahedron()
            {
                this.Edges = new Edge[13]; // 13 ... 4 tetrahedron where each 2 neighbouring share 3 edges 
            }

            /// <summary>
            /// Clear values.
            /// </summary>
            public void Clear()
            {
                Array.Clear(this.Edges, 0, this.Edges.Length);
            }
        }

        /// <summary>
        /// Edge of the cube or tetrahedron.
        /// </summary>
        public struct DEdge : IEquatable<DEdge>
        {
            int p1, p2;

            /// <summary>
            /// Constructor.
            /// </summary>
            /// <param name="x"></param>
            /// <param name="y"></param>
            public DEdge(int x, int y)
            {
                this.p1 = x;
                this.p2 = y;
            }

            /// <summary>
            /// Get hash code.
            /// </summary>
            /// <returns>hash code</returns>
            public override int GetHashCode()
            {
                return p1 * p2;
            }

            #region IEquatable<Edge> Members

            /// <summary>
            /// Define equals.
            /// </summary>
            /// <param name="other"></param>
            /// <returns>true if equals</returns>
            public bool Equals(DEdge other)
            {
                if ((this.p1 == other.p1) && (this.p2 == other.p2) || (this.p1 == other.p2) && (this.p2 == other.p1))
                    return (true);
                else
                    return (false);
            }

            #endregion
        }

        /// <summary>
        /// Bit valuation tetrahedron vertices.
        /// </summary>
        private struct Flags
        {
            /// <summary>
            /// Value indicate which edges of tetrahedron are intersect.
            /// </summary>
            public byte Value;

            public bool Set0 { set { this.Value |= 1; } } // D
            public bool Set1 { set { this.Value |= 2; } } // C
            public bool Set2 { set { this.Value |= 4; } } // B
            public bool Set3 { set { this.Value |= 8; } } // A
        }

        /// <summary>
        /// Structure used for storage tetrahedron values.
        /// </summary>
        /// <remarks>
        ///    C
        ///  / | \
        /// A -+- D
        ///  \ | /
        ///    B
        /// </remarks>
        private struct Tetrahedron
        {
            /// <summary>
            /// Values of tetrahedron.
            /// </summary>
            public double A, B, C, D;

            /// <summary>
            /// indexy pro hash funkci
            /// </summary>
            public int ia, ib, ic, id;

            /// <summary>
            /// Edge indices.
            /// </summary>
            public byte[] EdgeIndices;

            /// <summary>
            /// Define if vertices are same.
            /// </summary>
            public bool Useless;

            /// <summary>
            /// Attribute of intersect edges.
            /// </summary>
            public Flags Flags;

            /// <summary>
            /// Set indexs.
            /// </summary>
            public void SetIndex(int ia, int ib, int ic, int id)
            {
                this.ia = ia;
                this.ib = ib;
                this.ic = ic;
                this.id = id;
            }

            /// <summary>
            /// Values in vertices.
            /// </summary>
            public void SetValues(double a, double b, double c, double d)
            {
                if (a == b && b == c && c == d)
                    this.Useless = true;
                else
                    this.Useless = false;

                this.A = a;
                this.B = b;
                this.C = c;
                this.D = d;

                this.Flags = new Flags();
            }

            /// <summary>
            /// Setting edge indices of octahedron which define tetrahedron. 
            /// </summary>
            /// <param name="indices">edege indices</param>
            public void SetIndices(byte[] indices)
            {
                this.EdgeIndices = indices;
            }

            /// <summary>
            /// Attribute of intersect edges.
            /// </summary>
            /// <returns>TRUE if no one edege is intersect</returns>
            public bool Evaluate()
            {
                if (this.A < CCL.TresholdValue) this.Flags.Set3 = true;
                if (this.B < CCL.TresholdValue) this.Flags.Set2 = true;
                if (this.C < CCL.TresholdValue) this.Flags.Set1 = true;
                if (this.D < CCL.TresholdValue) this.Flags.Set0 = true;

                return (Flags.Value > 0) && (Flags.Value < 15);
            }
        }

        /// <summary>
        /// Structure define kd-tree node.
        /// </summary>
        private struct KdNode
        {
            public int pKdCell; // index to array of kdCells
            public double Min;   // minimum vol. value of cell
            public double Max;   // maximum vol. value of cell

            public KdNode(int pKdCell, double min, double max)
            {
                this.pKdCell = pKdCell;
                this.Min = min;
                this.Max = max;
            }
        }

        /// <summary>
        /// Structure contain cell data.
        /// </summary>
        private struct KdCell
        {
            public Cell cell;       // values in the conrners
            public Centers centers; // 6 neighbouring centres
            public int i, j, k;     // indexs in the grid .. used for determ. position
            public Normals normals; // normals

            public KdCell(Cell cell, Centers centers, Normals normals, int i, int j, int k)
            {
                this.cell = cell;
                this.centers = centers;
                this.normals = normals;
                this.i = i;
                this.j = j;
                this.k = k;
            }

            /// <summary>
            /// Minimum value from cell values.
            /// </summary>
            /// <returns>minimum value</returns>
            public double GetMin()
            {
                double min = cell.A;

                if (this.cell.B < min) min = cell.B;
                if (this.cell.C < min) min = cell.C;
                if (this.cell.D < min) min = cell.D;
                if (this.cell.E < min) min = cell.E;
                if (this.cell.F < min) min = cell.F;
                if (this.cell.G < min) min = cell.G;
                if (this.cell.H < min) min = cell.H;

                if (this.centers.Back < min) min = centers.Back;
                if (this.centers.Front < min) min = centers.Front;

                if (this.centers.Left < min) min = centers.Left;
                if (this.centers.Right < min) min = centers.Right;

                if (this.centers.Top < min) min = centers.Top;
                if (this.centers.Down < min) min = centers.Down;

                return min;
            }

            /// <summary>
            /// Maximum values from cell values.
            /// </summary>
            /// <returns>maximum values</returns>
            public double GetMax()
            {
                double max = cell.A;

                if (this.cell.B > max) max = cell.B;
                if (this.cell.C > max) max = cell.C;
                if (this.cell.D > max) max = cell.D;
                if (this.cell.E > max) max = cell.E;
                if (this.cell.F > max) max = cell.F;
                if (this.cell.G > max) max = cell.G;
                if (this.cell.H > max) max = cell.H;

                if (this.centers.Back > max) max = centers.Back;
                if (this.centers.Front > max) max = centers.Front;

                if (this.centers.Left > max) max = centers.Left;
                if (this.centers.Right > max) max = centers.Right;

                if (this.centers.Top > max) max = centers.Top;
                if (this.centers.Down > max) max = centers.Down;

                return max;
            }
        }

        /// <summary>
        /// Class used for sorting kd-tree nodes by minimum.
        /// </summary>
        private class SortByMin : IComparer<KdNode>
        {
            public int Compare(KdNode x, KdNode y)
            {
                double o1 = x.Min;
                double o2 = y.Min;

                if (o1 > o2)
                    return 1;

                if (o1 < o2)
                    return -1;
                else
                    return 0;
            }
        }

        /// <summary>
        /// Class used for sorting kd-tree nodes by maximum.
        /// </summary>
        private class SortByMax : IComparer<KdNode>
        {
            public int Compare(KdNode x, KdNode y)
            {
                double o1 = x.Max;
                double o2 = y.Max;

                if (o1 > o2)
                    return 1;

                if (o1 < o2)
                    return -1;
                else
                    return 0;
            }
        }

        #endregion
    }
}
