using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms.Design;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using Zcu.Mve.Strings;


namespace Zcu.Gsvd.Extractions
{
    /// <summary>
    /// Module load volumetric data into RegGridND structure. Data are load from .vol file.
    /// For successfull data reading is necessary .vif which should be in the same directory and name. 
    /// </summary>
    [ModuleInfo("Novak Frantisek", "Module load volumetric data from .vol file format.", 
        IconName = "Default.ico" ,Assemblied = "2006-07-22", Category = "Gsvd")]
    [PortInfo("Out", "RegGridND")]
    public class VolLoader : Zcu.Mve.Core.Module
    {
        /// <summary>
        /// Volumetric data file.
        /// </summary>
        private string volFileName = string.Empty;

        /// <summary>
        /// Volumetric data file description.
        /// </summary>
        private string vifFileName = string.Empty;

        private bool paramsChanged = true;
        private string pointAttributName = "Values";

        private RegGridND grid3D = null;

        /// <summary>
        /// Create out port.
        /// </summary>
        public VolLoader()
        {
            AddOutPort("Out", typeof(RegGridND));
            AddInPort("FileName", typeof(MveString), false);
        }

        /// <summary>
        /// Entry point of module. Call method for reading data and sets data on out port.
        /// </summary>
        public override void Execute()
        {
            if (!IsAllDataSame() || paramsChanged)
            {
                if (this.GetInput("FileName") != null)
                    this.volFileName = (MveString)this.GetInput("FileName");

                if (System.IO.File.Exists(this.volFileName))
                {
                    System.GC.Collect();

                    LoadData();

                    this.SetOutput("Out", this.grid3D);
                    this.paramsChanged = false;
                }
                else
                    throw new FileNotFoundException(string.Format("File {0} not exists. File name cannot be set to an empty string!", this.volFileName));
            }
            else            
                this.SetOutput("Out", this.grid3D, true);              
        }

        /// <summary>
        /// Loading vol. data.
        /// </summary>
        private void LoadData()
        {
            if (this.volFileName.EndsWith(".vol") || this.volFileName.EndsWith(".VOL"))
            {
                this.vifFileName = this.volFileName.ToLower().Replace(".vol", ".vif");
                
                if (System.IO.File.Exists(this.vifFileName))
                {
                    int ticks = Environment.TickCount;
                    StreamReader sr = new StreamReader(this.vifFileName, Encoding.ASCII);
                    // load data description from .vif file
                    Vif vif = new Vif(sr.ReadToEnd(), this.vifFileName);

                    ProgressInfoChangedEventHandler bar = new ProgressInfoChangedEventHandler(UpdateProgress);
                    
                    FileStream fs = new FileStream(this.volFileName, FileMode.Open, FileAccess.Read, FileShare.None);
                    BinaryReader br = new BinaryReader(fs);
                    // load volumetric data
                    Vol vol = new Vol(br, vif, bar);

                    this.grid3D = new RegGridND((int) (vif.Resolution.X - 1), (int) (vif.Resolution.Y - 1), (int) (vif.Resolution.Z - 1));
                    this.grid3D.AddPointAttrs(this.pointAttributName, vol.Data);

                    grid3D.UserTags.Add("Source", vif.Source);
                    grid3D.UserTags.Add("Caption", vif.Caption);
                    grid3D.UserTags.Add("Min", vol.Min);
                    grid3D.UserTags.Add("Max", vol.Max);
                    grid3D.UserTags.Add("Position", vif.Position);
                    grid3D.UserTags.Add("Rotation", vif.Rotation);
                    grid3D.UserTags.Add("Distortion", vif.Distortion);
                    
                    FileInfo fi = new FileInfo(this.volFileName);
                    System.Diagnostics.Debug.WriteLine(string.Format("File {0} loading time: {1} ms", fi.Name, Environment.TickCount - ticks));
                }
                else
                    throw new FileNotFoundException(string.Format("File \"{0}\" not found. Vif file is contain description of .vol file ", this.vifFileName));
            }
            else
                throw new FileNotFoundException(string.Format("Bad file \"{0}\" format, file should be .vol format.", this.volFileName));
        }

        /// <summary>
        /// Change progress bar value.
        /// </summary>
        private void UpdateProgress(object sender, double percent)
        {
            ProgressInfo(percent);
        }

        #region Properties

        /// <summary>
        /// File to load data from. Should be in the .obj format. The property has no effect when there's a connection at the Filename port
        /// </summary>
        [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        [Description("File to load data from. Should be in the .vol format.")]
        public string FileName
        {
            get { return (this.volFileName); }
            set
            {
                if (value != "")
                {
                    volFileName = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "FileName property cannot be set to an empty string!");
                }
            }
        }

        /// <summary>
        /// Set name of gridND point attributes. Default set to Values.
        /// </summary>
        [Description("Set name of gridND point attributes. Default set to \"Values\"")] 
        public string PointAttributeName
        {
            set 
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributName; }
        }

        #endregion
    }
}
