using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Gsvd.Extractions
{
    /// <summary>
    /// Module extracts iso-surface from input volumetric data with Marching Cubes algorithm. 
    /// Generates large amount of triangles/points to render - it would be nice to use some module 
    /// for triangle mesh simplification or another suitable post-processing.
    /// Module shall be used with 'VolLoader' module, which provides desired input data. 
    /// </summary>
    [ModuleInfo("Ji Zapletal", "Marching Cubes iso-surface extractor.\n\nGenerates large amount of " +
        "triangles to render - use some module for triangle mesh simplification or another suitable " +
        "post-processing.\nModule shall be used with 'VolLoader' module, which provides desired input data.",
        IconName = "MC.ico",
        Category = "Gsvd")]
    [PortInfo("InThreshold", "This value defines iso-surface threshold (in percents).\nNon-required input port.")]
    [PortInfo("InRegGridND", "Input volumetric data.")]
    [PortInfo("Out", "Output data - triangle mesh.")]
    public class MarchingCubes : Zcu.Mve.Core.Module
    {
        private bool paramsChanged = true, linearInterpolation = true;
        private int cubeIndex = 0;
        private double min = double.MaxValue, max = double.MinValue;
        private int[] size;
        private double isoValue = 0, threshold = 50;
        private string pointAttributesName = "Values";
        private string normalAttributeName = "Normal3D";
        private bool useThresholdAsIsoValue = false;

        private Point3D distortion;
        private Cube cube, previousCube; // actual and previous cube
        private Vertex[] vertList = null; // store generated vertices
        private List<Vertex> pts = null;
        private List<Triangle> indices = null;
        private int hashIndex = 0;
        private int[] indexList = null;

        private RegGridND inData = null; // input data
        private MveArray data = null;  // volumetric data
        private TriangleMesh mesh = null; // output data
        private UniformDataArray points, normals, triangles; // arrays of primitives for output

        private Dictionary<Edge, int> dict;
        private Edge edge = new Edge();

        /// <summary>
        /// Constructor. Creates input and output ports.
        /// </summary>
        public MarchingCubes()
        {
            AddInPort("InThreshold", typeof(Integer), false);
            AddInPort("InRegGridND", typeof(RegGridND));
            AddOutPort("Out", typeof(TriangleMesh));
        } // of MarchingCubes

        /// <summary>
        /// Executes module functionality.
        /// </summary>
        public override void Execute()
        {
            this.ProgressInfo(0f);
            if (!IsAllDataSame() || this.paramsChanged) // if I have to compute again 
            {
                ComputeIsoSurface();
                this.SetOutput("Out", this.mesh);
            }
            else
                this.SetOutput("Out", this.mesh, true);
        } // of Execute

        /// <summary>
        /// This method initializes variables and call cube extracting method. 
        /// </summary>
        private void ComputeIsoSurface()
        {
            System.GC.Collect();
            Initialize();

            Console.Write("Marching Cubes\n Extracting iso-surfaces...");
            long ticks = Environment.TickCount;
            ExtractCubesFromDataGrid(); // extract cubes
            Console.WriteLine(" {0} ms", Environment.TickCount - ticks);

            Console.WriteLine("\n  Linear interp.: {0}\n  Threshold:      {1}%\n  Triangles:       {2}\n   Points:           {3}\n",
                linearInterpolation, threshold, indices.Count / 3, pts.Count);

            CreateOutput(); // creates from computed data output data
        } // ComputeIsoSurface

        /// <summary>
        /// Initializes variables.
        /// </summary>
        private void Initialize()
        {
            paramsChanged = false;
            inData = (RegGridND)this.GetInput("InRegGridND");
            data = inData.GetPointAttrs(this.pointAttributesName);
            size = data.Lengths;
            distortion = new Vector3D(1, 1, 1);
            previousCube = cube = new Cube(new Vertex[8]);
            vertList = new Vertex[12];
            pts = new List<Vertex>();
            indices = new List<Triangle>();
            indexList = new int[12];
            this.ProgressInfo(0f);

            bool allAtributes = true;

            if (inData.UserTags.ContainsKey("Min"))
                min = (Scalar)inData.UserTags["Min"];
            else
            {
                min = 0;
                allAtributes = false;
            }

            if (inData.UserTags.ContainsKey("Max"))
                max = (Scalar)inData.UserTags["Max"];
            else
            {
                max = 255;
                allAtributes = false;
            }

            if (inData.UserTags.ContainsKey("Distortion"))
                distortion = (Point3D)inData.UserTags["Distortion"];
            else
            {
                distortion = new Point3D(1, 1, 1);
                allAtributes = false;
            }

            if (!allAtributes)
                throw new Exception("Attributes not set. For extraction used default values.");

            if (this.GetInput("InThreshold") != null)
                threshold = (Integer)this.GetInput("InThreshold");

            if (useThresholdAsIsoValue)
                isoValue = threshold;
            else
                isoValue = min + ((max - min) * (threshold / 100f));

            dict = new Dictionary<Edge, int>(this.size[0] * this.size[1] * this.size[2]);
        }

        private void UpdateProgress(object sender, float percent)
        {
            ProgressInfo(percent);
        }

        /// <summary>
        /// This method makes cubes from values in two neghbouring slices. 
        /// The convention of cubes vertices numbering is shown below. 
        /// After each cube is created, it is commited to the next step - checked whether it 
        /// is intersected by searched iso-surface and eventually appropriate triangles are generated.
        /// 
        ///   4 ------ 5
        ///  /|       /|
        /// 7 ------ 6 |
        /// | |      | |
        /// | 0 -----| 1
        /// |/       |/
        /// 3 ------ 2
        /// </summary> 
        private void ExtractCubesFromDataGrid()
        {
            //ProgressInfoChangedEventHandler barHandler = new ProgressInfoChangedEventHandler(UpdateProgress);
            //barHandler(this, 0.0f);

            for (int i = 0; i < this.size[0] - 1; i++)
            {
                for (int j = 0; j < this.size[1] - 1; j++)
                {
                    for (int k = 0; k < this.size[2] - 1; k++)
                    {
                        if ((i > 0) && (j > 0) && (k > 0)) // take shared vertices form previous cube
                        {
                            cube.p[0] = previousCube.p[3];

                            cube.p[1] = previousCube.p[2];

                            cube.p[2].coords.X = i + 1; cube.p[2].coords.Y = j; cube.p[2].coords.Z = k + 1;
                            cube.p[2].value = (Scalar)data[i + 1, j, k + 1];
                            cube.p[2].index = i + 1 + (j * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[3].coords.X = i; cube.p[3].coords.Y = j; cube.p[3].coords.Z = k + 1;
                            cube.p[3].value = (Scalar)data[i, j, k + 1];
                            cube.p[3].index = i + (j * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[4] = previousCube.p[7];

                            cube.p[5] = previousCube.p[6];

                            cube.p[6].coords.X = i + 1; cube.p[6].coords.Y = j + 1; cube.p[6].coords.Z = k + 1;
                            cube.p[6].value = (Scalar)data[i + 1, j + 1, k + 1];
                            cube.p[6].index = i + 1 + ((j + 1) * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[7].coords.X = i; cube.p[7].coords.Y = j + 1; cube.p[7].coords.Z = k + 1;
                            cube.p[7].value = (Scalar)data[i, j + 1, k + 1];
                            cube.p[7].index = i + ((j + 1) * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);
                        }
                        else
                        {
                            // assign coordinates and values to each vertex of the cube
                            cube.p[0].coords.X = i; cube.p[0].coords.Y = j; cube.p[0].coords.Z = k;
                            cube.p[0].value = (Scalar)data[i, j, k];
                            cube.p[0].index = i + (j * this.size[0]) + (k * this.size[0] * this.size[1]);

                            cube.p[1].coords.X = i + 1; cube.p[1].coords.Y = j; cube.p[1].coords.Z = k;
                            cube.p[1].value = (Scalar)data[i + 1, j, k];
                            cube.p[1].index = i + 1 + (j * this.size[0]) + (k * this.size[0] * this.size[1]);

                            cube.p[2].coords.X = i + 1; cube.p[2].coords.Y = j; cube.p[2].coords.Z = k + 1;
                            cube.p[2].value = (Scalar)data[i + 1, j, k + 1];
                            cube.p[2].index = i + 1 + (j * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[3].coords.X = i; cube.p[3].coords.Y = j; cube.p[3].coords.Z = k + 1;
                            cube.p[3].value = (Scalar)data[i, j, k + 1];
                            cube.p[3].index = i + (j * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[4].coords.X = i; cube.p[4].coords.Y = j + 1; cube.p[4].coords.Z = k;
                            cube.p[4].value = (Scalar)data[i, j + 1, k];
                            cube.p[4].index = i + ((j + 1) * this.size[0]) + (k * this.size[0] * this.size[1]);

                            cube.p[5].coords.X = i + 1; cube.p[5].coords.Y = j + 1; cube.p[5].coords.Z = k;
                            cube.p[5].value = (Scalar)data[i + 1, j + 1, k];
                            cube.p[5].index = i + 1 + ((j + 1) * this.size[0]) + (k * this.size[0] * this.size[1]);

                            cube.p[6].coords.X = i + 1; cube.p[6].coords.Y = j + 1; cube.p[6].coords.Z = k + 1;
                            cube.p[6].value = (Scalar)data[i + 1, j + 1, k + 1];
                            cube.p[6].index = i + 1 + ((j + 1) * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);

                            cube.p[7].coords.X = i; cube.p[7].coords.Y = j + 1; cube.p[7].coords.Z = k + 1;
                            cube.p[7].value = (Scalar)data[i, j + 1, k + 1];
                            cube.p[7].index = i + ((j + 1) * this.size[0]) + ((k + 1) * this.size[0] * this.size[1]);
                        }

                        GenerateTriangles();
                    }
                }
                this.ProgressInfo(((float)i / this.size[0]) * 0.85f);
                //dict = new Dictionary<Edge, int>(this.size[1] * this.size[2]);
            }
        } // of ExtractCubesFromDataGrid

        /// <summary>
        /// Checks whether a cube is intersected by searched iso-surface and eventually 
        /// generates appropriate triangles.
        /// </summary>
        private void GenerateTriangles()
        {
            cubeIndex = 0;

            // create index which vertices of the cube are under iso-surface
            if (cube.p[0].value < isoValue) cubeIndex |= 1;
            if (cube.p[1].value < isoValue) cubeIndex |= 2;
            if (cube.p[2].value < isoValue) cubeIndex |= 4;
            if (cube.p[3].value < isoValue) cubeIndex |= 8;
            if (cube.p[4].value < isoValue) cubeIndex |= 16;
            if (cube.p[5].value < isoValue) cubeIndex |= 32;
            if (cube.p[6].value < isoValue) cubeIndex |= 64;
            if (cube.p[7].value < isoValue) cubeIndex |= 128;

            // stop if there is no intersection 
            if (Misc.edgeTable[cubeIndex] == 0)
                return;

            // calculate normals for all cubes vertices
            for (int i = 0; i < 8; i++)
                cube.p[i].normal = Misc.CalcNormal(cube.p[i], data, size);

            previousCube = cube;

            // find edges where we found intersection and calculate positions of triangles vertices
            if ((Misc.edgeTable[cubeIndex] & 1) != 0)
                GetIndex(0, cube.p[0], cube.p[1]);
            if ((Misc.edgeTable[cubeIndex] & 2) != 0)
                GetIndex(1, cube.p[1], cube.p[2]);
            if ((Misc.edgeTable[cubeIndex] & 4) != 0)
                GetIndex(2, cube.p[2], cube.p[3]);
            if ((Misc.edgeTable[cubeIndex] & 8) != 0)
                GetIndex(3, cube.p[3], cube.p[0]);
            if ((Misc.edgeTable[cubeIndex] & 16) != 0)
                GetIndex(4, cube.p[4], cube.p[5]);
            if ((Misc.edgeTable[cubeIndex] & 32) != 0)
                GetIndex(5, cube.p[5], cube.p[6]);
            if ((Misc.edgeTable[cubeIndex] & 64) != 0)
                GetIndex(6, cube.p[6], cube.p[7]);
            if ((Misc.edgeTable[cubeIndex] & 128) != 0)
                GetIndex(7, cube.p[7], cube.p[4]);
            if ((Misc.edgeTable[cubeIndex] & 256) != 0)
                GetIndex(8, cube.p[0], cube.p[4]);
            if ((Misc.edgeTable[cubeIndex] & 512) != 0)
                GetIndex(9, cube.p[1], cube.p[5]);
            if ((Misc.edgeTable[cubeIndex] & 1024) != 0)
                GetIndex(10, cube.p[2], cube.p[6]);
            if ((Misc.edgeTable[cubeIndex] & 2048) != 0)
                GetIndex(11, cube.p[3], cube.p[7]);

            // from calculated vertices create triangles
            for (int i = 0; Misc.triTable[cubeIndex, i] != -1; i += 3)
            {
                indices.Add(new Triangle(
                    indexList[Misc.triTable[cubeIndex, i]],
                    indexList[Misc.triTable[cubeIndex, i + 1]],
                    indexList[Misc.triTable[cubeIndex, i + 2]]
                    ));
            }
        } // of GenerateTriangles

        /// <summary>
        /// Gets index of the point form array of already computed points or adds new point into this array.
        /// </summary>
        /// <param name="index">Index of intersected edge.</param>
        /// <param name="a">First vertex of the edge.</param>
        /// <param name="b">Second vertex of the edge.</param>
        private void GetIndex(int index, Vertex a, Vertex b)
        {
            if (a.value < b.value)
            {
                edge.p1 = a.index;
                edge.p2 = b.index;
            }
            else
            {
                edge.p1 = b.index;
                edge.p2 = a.index;
            }

            indexList[index] = FindPointIndex(edge);

            if (indexList[index] == -1)
            {
                vertList[index] = Misc.Interpolate(a, b, inData, isoValue, distortion, linearInterpolation);
                dict.Add(edge, hashIndex);
                pts.Add(vertList[index]);
                indexList[index] = hashIndex;
                hashIndex++;
            }
        } // of GetIndex

        /// <summary>
        /// Gets index = value from the hashtable of indices.
        /// </summary>
        /// <param name="edge">On which edge the point is.</param>
        /// <returns>Index of the point.</returns>
        private int FindPointIndex(Edge edge)
        {
            int value;
            if (dict.TryGetValue(edge, out value))
                return value;
            else
                return -1;
        } // of FindPointIndex

        /// <summary>
        ///  Creates from computed data output data
        /// </summary>
        private void CreateOutput()
        {
            mesh = new TriangleMesh();
            points = new UniformDataArray(typeof(Point3D), pts.Count);
            for (int i = 0; i < pts.Count; i++)
                points[i] = new Point3D(pts[i].coords.X, pts[i].coords.Y, pts[i].coords.Z);
            mesh.Points = points;
            points = null;
            this.ProgressInfo(0.9f);

            normals = new UniformDataArray(typeof(Normal3D), pts.Count);
            for (int i = 0; i < pts.Count; i++)
                normals[i] = new Normal3D(pts[i].normal.X, pts[i].normal.Y, pts[i].normal.Z);
            mesh.AddPointAttrs(normalAttributeName, normals);
            normals = null;
            this.ProgressInfo(0.95f);

            triangles = new UniformDataArray(typeof(Triangle), indices.Count);
            for (int i = 0; i < indices.Count; i++)
                triangles[i] = indices[i];

            mesh.AddCells("Triangles", triangles);
            triangles = null;
            indices = null;
            vertList = null;
            inData = null;
            pts = null;
            indexList = null;
            hashIndex = 0;
            data = null;
            dict = null;
            System.GC.Collect();
            this.ProgressInfo(1f);
        } // of CreateOutput

        #region Properties

        /// <summary>
        /// Property for threshold variable.
        /// </summary>
        [Description("Threshold of iso-surface value in percents. \nIf 'useThresholdAsIsoValue' flag is set to True,"
        + " is this value directly used as iso value to be searched (used for implicit surfaces visualization).")]
        public double Threshold
        {
            get { return (this.threshold); }
            set
            {
                if ((value >= 0) && (value <= 100))
                {
                    this.threshold = value;
                    this.paramsChanged = true;
                }
                else
                {
                    this.threshold = 50;
                    this.paramsChanged = true;
                }
            }
        }

        /// <summary>
        /// Property for linearInterpolation variable.
        /// </summary>
        [Description("Interpolation type:\n  true = linear interpolation\n  false = center point.")]
        public bool LinearInterpolation
        {
            get { return (this.linearInterpolation); }
            set
            {
                this.linearInterpolation = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Set name of RegGridND point attributes. Default set to Values.
        /// </summary>
        [Description("Set name of RegGridND point attributes. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND normal attribute. Default set to Values.
        /// </summary>
        [Description("Set name of normals point attribute. Default set to \"Normal3D\"")]
        public string NormalAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.normalAttributeName = value;
                    this.paramsChanged = true;
                }
            }
            get { return this.normalAttributeName; }
        }
        #endregion

        /// <summary>
        /// Property for linearInterpolation variable.
        /// </summary>
        [Description("Specifies whether to use input threshold as percentual value from min/max interval or directly as iso value.")]
        public bool UseThresholdAsIsoValue
        {
            get { return (this.useThresholdAsIsoValue); }
            set
            {
                this.useThresholdAsIsoValue = value;
                this.paramsChanged = true;
            }
        }

    } // of MarchingCubes class
} // of namespace Zcu.Gsvd.Extractions
