using System;
using System.IO;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Gsvd.Extractions
{
    /// <summary>
    /// Read volumetric data from .vol file.
    /// </summary>
    internal class Vol
    {
        /// <summary>
        /// Minimum value of loaded volumetric data.
        /// </summary>
        private double min;

        /// <summary>
        /// Maximum value of loaded volumetric data.
        /// </summary>
        private double max;

        /// <summary>
        /// Volumetric data.
        /// </summary>
        private MveArray data = null;

        /// <summary>
        /// Handle to progress bar
        /// </summary>
        private VolLoader.ProgressInfoChangedEventHandler barHandler = null;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="br">file data source</param>
        /// <param name="vif">description of .vol file</param>
        /// <param name="barHandler">handle to progress bar</param>
        public Vol(BinaryReader br, Vif vif, VolLoader.ProgressInfoChangedEventHandler barHandler)
        {
            this.barHandler = barHandler;

            this.data = new MveArray(typeof(Scalar), (int) vif.Resolution.X, (int) vif.Resolution.Y, (int) vif.Resolution.Z);

            br.ReadBytes(vif.HeaderSize); // skip header

            this.barHandler(this, 0.0f); 

            switch (vif.BytesPerVoxel)
            {
                case 1:
                    LoadAsByte(br);
                    break;

                case 2:
                    LoadAsShort(br);
                    break;

                case 4:
                    LoadAsInt(br);
                    break;

                default :
                    throw new FormatException("Unsupported volumetric data type.");
            }
        }

        #region Data load

        /// <summary>
        /// Method load data as byte. This mean data have size 1 bytes per voxel.
        /// </summary>
        /// <param name="br">volumetric data source</param>
        private void LoadAsByte(BinaryReader br)
        {
            byte h;
            bool first = true;

            for (int k = 0; k < this.data.Lengths[2]; k++)
            {
                for (int j = 0; j < this.data.Lengths[1]; j++)
                {
                    for (int i = 0; i < this.data.Lengths[0]; i++)
                    {
                        if (first)
                        {
                            this.min = this.max = h = br.ReadByte();
                            first = false;
                        }
                        else
                        {
                            h = br.ReadByte();
                            this.min = this.min > h ? h : this.min;
                            this.max = this.max < h ? h : this.max;
                        }
                        data[i, j, k] = (Scalar)h;
                    }
                }

                this.barHandler.Invoke(this, ((float)k / (float)this.data.Lengths[2]));
            }
        }
  
    

        /// <summary>
        /// Method load data as short. This mean data have size 2 bytes per voxel.
        /// </summary>
        /// <param name="br">volumetric data source</param>
        private void LoadAsShort(BinaryReader br)
        {
            bool first = true;
            short h;

            for (int k = 0; k < this.data.Lengths[2]; k++)
            {
                for (int j = 0; j < this.data.Lengths[1]; j++)
                {
                    for (int i = 0; i < this.data.Lengths[0]; i++)
                    {
                        if (first)
                        {
                            this.min = this.max = h = br.ReadInt16();
                            first = false;
                        }
                        else
                        {
                            h = br.ReadInt16();
                            this.min = this.min > h ? h : this.min;
                            this.max = this.max < h ? h : this.max;
                        }
                        data[i, j, k] = (Scalar)h;
                    }
                }

                this.barHandler.Invoke(this, ((float)k / (float)this.data.Lengths[2]));
            }

        }

        /// <summary>
        /// Method load data as int. This mean data have size 4 bytes per voxel.
        /// </summary>
        /// <param name="br">volumetric data source</param>
        private void LoadAsInt(BinaryReader br)
        {
            int h;
            bool first = true;

            for (int k = 0; k < this.data.Lengths[2]; k++)
            {
                for (int j = 0; j < this.data.Lengths[1]; j++)
                {
                    for (int i = 0; i < this.data.Lengths[0]; i++)
                    {
                        if (first)
                        {
                            this.min = this.max = h = br.ReadInt16();
                            first = false;
                        }
                        else
                        {
                            h = br.ReadInt32();
                            this.min = this.min > h ? h : this.min;
                            this.max = this.max < h ? h : this.max;
                        }
                        data[i, j, k] = (Scalar)h;
                    }
                }

                this.barHandler.Invoke(this, ((float)k / (float)this.data.Lengths[2]));
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Loaded volumetric data.
        /// </summary>
        public MveArray Data
        {
            get { return this.data; }
        }

        /// <summary>
        /// Get maximal value in vol. data.
        /// </summary>
        public Scalar Max
        {
            get { return this.max; }
        }

        /// <summary>
        /// Get minimal value in vol. data.
        /// </summary>
        public Scalar Min
        {
            get { return this.min; }
        }

        #endregion
    }
}
